/* $Id: acre_main.c,v 1.1 2004/04/30 05:17:24 Joel M. Malard Exp d3j540 $ */

static char help[] =
"ACRE estimates (co)variance components for a linear model of the \n\
form Ax=b+e where A is a dense matrix and the covariance matrix \n\
of e has the form L*Transpose(L) where L is lower triangular and \n\
known up to the position of its non-zero elements\n\n\
The input arguments are:\n\n\
       -acre_ascii : all input files are either binary or ascii.\n\n\
       -acre_exact <tolerance> : controls the stopping criteria of the TAO\n\
       solver, that is blmvm.\n\n\
       -acre_fA <filename> : the name of the file containing the dense\n\
       exogenous matrix A. The matrix A is of type MATMPIDENSE (PETSc).\n\n\
       -acre_fb <filename> : the name of the file containing the response\n\
       vector.\n\n\
       -acre_fsL <filename> : the name of the file containing the sparsity\n\
       structure of L. The structural file is of type MATMPIAIJ (PETSc)\n\n\
       -acre_flb <filename> : the name of an optional file containing\n\
       lower bounds on the parameters.\n\n\
       -acre_fub <filename> : the name of an optional file containing\n\
       upper bounds on the parameters.\n\n\
       -acre_fx <filename> : the name of an optional file containing an\n\
       initial guess of the parameters.\n\n\
       -acre_g [cramer|adjoint] : the type of gradient formulation. Cramer\n\
       is the default. The (mixed tangent-)adjoint formulation is typically\n\
       slower; it is not supported when -acre_xc is greater than one.\n\n\
       -acre_help : print this message and exit.\n\n\
       -acre_i <number> : maximum number of non-linear iteration steps.\n\n\
       -acre_ml : signals that the ML estimates are to be needed instead\n\
       of the default REML estimates.\n\n\
       -acre_vec <number> : number of vectors used for quasi-Newton.\n\n\
       -acre_xc <number> : the number of columns in the 2D-processor mesh.\n\
       The matrix A is distributed along that 2D mesh. The sparse matrix L\n\
       and its inverse are replicated across all process columns.\n\n";

/* include files */
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "tao.h"
#include "acre.h"
#include "petscis.h"   
#include "petscmat.h" 
#include <unistd.h>
#include "copyright.h"

#define APP_BAND	0
#define APP_BLOCK	1

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **args)
{
	TAO_SOLVER reml;
	TAO_APPLICATION reml_app;
	acre_cntxt cntxt;

	Vec	b;		/* response vector */
	Vec	x;		/* vector of parameters */
	Vec	cVec;		/* Variance Components */
	Vec	lVec;		/* lower bounds on variance components */
	Vec	uVec;		/* upper bounds on variance components */
	Mat	A, AT;          /* dense exogenous matrix */
	Mat	sL;		/* structure of the matrix L */

	int	n = 1;		/* number of global columns of A */
	int	m = 10;		/* number of global rows of A */
	int	nq = 1;      /* number of variance components */
	int	maxits = 50;	/* maximum number of REML/ML iterations */
	char	gtype[8];	/* gradient type */
	char	ftype[8];	/* likelihood type */
	char	infile[3000];	/* input file name */
	int	w_rank, w_size;	/* global process rank and communicator size*/
	int	c_rank, c_size;	/* column process rank and communicator size */
	int	r_rank, r_size;	/* row process rank and communicator size */
	int	xc=1;
	int	nv = 5;

	int	igrad = ACRE_CRAMER_GRADIENT;

	/** 
	 ** Other declarations
	 **/
	PetscScalar  one = 1.0;
	PetscScalar  res;
	
	int mm;

	int     nrows, *rows, *cols;
	int     i, ierr;
	int     start, end;
	PetscTruth format, flg;
	MPI_Comm qcomm, rcomm;

	/***
	 *** Initialize MPI and split PETSC_COMM_WORLD.
	 ***/
	ierr = MPI_Init(&argc, &args);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank ( MPI_COMM_WORLD, &w_rank ); 
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_size ( MPI_COMM_WORLD, &w_size ); 
		assert(ierr==MPI_SUCCESS);

	/***
	 ***  Split MPI_COMM_WORLD if xc has a value greater than one.
	 ***
	 *** Argument -acre_xc is the number of processors among
	 *** which every row of the exogenous matrix A is
	 *** distributed.
	 ***  
	 ***/

	xc = 1;
	for ( i=0; i<argc; i++ )
	{
		if ( strncasecmp(args[i],"-acre_xc", 8)==0 )
		{
			xc = atoi(args[i+1]);
			/* Define the processor columns and reset
			 * the PETSC_COMM_WORLD communicator.
			 */
			if ( w_size%xc!=0 )
			{
				if(w_rank==0)
					PetscPrintf(PETSC_COMM_SELF,
					"%s %d %s %s %d\n",
					"The number of process groups", xc,
					"is not an integral divisor of",
					"the number of processors",
					w_size);
				ierr = TaoFinalize();		CHKERRQ(ierr);
				exit(0);
			}
		}
	}

	/* It would be nice to be able to use this call but it has been
	 * known to cause problems with MPI_Bcast; a local variant of
	 * this call would be fine.
	 */

	/*ierr = PetscOptionsGetInt( PETSC_NULL,
	 * "-acre_xc", &xc, &flg);CHKERRQ(ierr);
	 */


	if ( xc!=1 )
	{
		/* Define the processor columns and reset
		 * the PETSC_COMM_WORLD communicator.
		 */
		if ( w_size%xc!=0 )
		{
			if(w_rank==0)
				PetscPrintf(PETSC_COMM_SELF,
				"%s %d %s %s %d\n",
				"The number of process groups", xc,
				"is not an integral divisor of",
				"the number of processors",
				w_size);
			exit(0);
		}
		ierr = MPI_Comm_split(MPI_COMM_WORLD,
			w_rank%xc, w_rank/xc, &qcomm);
			assert(ierr==MPI_SUCCESS);

		ierr = PetscSetCommWorld(qcomm);		CHKERRQ(ierr);

		ierr = MPI_Comm_rank ( qcomm, &c_rank ); 
			assert(ierr==MPI_SUCCESS);

		ierr = MPI_Comm_size ( qcomm, &c_size ); 
			assert(ierr==MPI_SUCCESS);

		/* Define the processor rows. */
		ierr = MPI_Comm_split(MPI_COMM_WORLD, w_rank/xc,
			w_rank%xc, &rcomm);
			assert(ierr==MPI_SUCCESS);
			
		ierr = MPI_Comm_rank ( rcomm, &r_rank ); 
			assert(ierr==MPI_SUCCESS);

		ierr = MPI_Comm_size ( rcomm, &r_size ); 
			assert(ierr==MPI_SUCCESS);
	}
	else
	{
		c_rank = w_rank;	r_rank=0;
		c_size = w_size;	r_size=1;
	}

	/***
	 ***  Start up PETSC and find out the command line arguments other
	 ***  than -acre_xc.
	 ***/
	TaoInitialize(&argc,&args,(char *)0,help);

	/**** Check whether or not to print the copyright notice. 
	 ****/
	ierr = PetscOptionsHasName(PETSC_NULL, "-acre_silent",
		&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
	  PetscPrintf(PETSC_COMM_SELF,
	  "\n%s%s\n%s\n%s%s%s%s\n%s%s%s%s\n%s%s%s\n",
	  "Copyright 1997 University of Edinburgh\n",
	  "Copyright 2002, 2004 Battelle Memorial Institute\n",
	  "This file is part of the ACRE version 1.1.\n",
	  "ACRE is free software; you can redistribute it and/or\n",
	  "modify it under the terms of the GNU General Public License\n",
	  "as published by the Free Software Foundation; either version 2\n",
	  "of the License, or (at your option) any later version.\n",
	  "ACRE is distributed in the hope that it will be useful,\n",
	  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n",
	  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n",
	  "GNU General Public License for more details.\n",
	  "You should have received a copy of the GNU General Public License\n",
	  "along with ACRE; if not, write to the Free Software Foundation,\n",
	  "Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA\n");
	}

	/**** Check if the input files are ASCII or binary. At present only
	 **** binary input is supported for matrices.
	 ****/
	ierr = PetscOptionsHasName(PETSC_NULL, "-acre_ascii",
		&format);					CHKERRQ(ierr);

	/**** Argument -acre_fA gives the name of the binary file
	 **** that contains the exogenous matrix A. Each process
	 **** column read the entire matrix and then extracts the
	 **** appropriate columns.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fA", infile,
		3000,&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fA is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		ierr = MPI_Finalize();
			assert(ierr==MPI_SUCCESS);
		exit(0);
	}

	if ( xc==1 )
	{
		ierr = acre_read_dense_matrix(PETSC_COMM_WORLD,
			infile, format, &A);			CHKERRQ(ierr);
		ierr = MatGetSize(A,&m, &n);			CHKERRQ(ierr);
	}
	else
	{
		int	rn;	/* number of local columns of A */
		int	rs;	/* index of first local columns of A */
		int	re;	/* upper bound for local columns indices */

		ierr = acre_read_dense_matrix(PETSC_COMM_WORLD,
			infile, format, &AT);			CHKERRQ(ierr);

		ierr = MatGetSize(AT,&m, &n);			CHKERRQ(ierr);
		if(n>m)
		{
			if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
			"Exogenous Matrix A has fewer columns than rows\n");
			ierr = TaoFinalize();			CHKERRQ(ierr);
			ierr = MPI_Finalize();
				assert(ierr==MPI_SUCCESS);
			exit(0);
		}
		/***
		 *** Find which columns of the global exogenous
		 *** matrix belong to the current PETSC_COMM_WORLD
		 *** and compute an appropriately skinned A.
		 ***/
	
		rs = r_rank*(n/xc);
		rs += ( r_rank < n%xc ) ? r_rank : (n%xc);

		re = (r_rank+1)*(n/xc);
		re += ( (r_rank+1) < n%xc ) ? (r_rank+1) : (n%xc);
		if ( r_rank+1==xc ) re = n;
		rn = re - rs;

		ierr = MatGetOwnershipRange(AT,&start,&end);	CHKERRQ(ierr);
		nrows = end - start;

		ierr = MatCreateMPIDense(PETSC_COMM_WORLD,
			PETSC_DECIDE, PETSC_DECIDE, m, rn,
			TAO_NULL, &A);				CHKERRQ(ierr);
		if ( nrows!=0 && rn!=0 )
		{
			PetscScalar *vals;
			ierr = PetscMalloc(nrows*sizeof(int),
				&rows);				CHKERRQ(ierr);
			for ( i=0;i<nrows; i++ ) rows[i] = start+i;
			ierr = PetscMalloc(rn*sizeof(int),
				&cols);				CHKERRQ(ierr);
			for ( i=0;i<rn; i++ ) cols[i] = i;
			ierr = MatGetArray(AT,&vals);		CHKERRQ(ierr);
			ierr = MatSetValues( A, nrows, rows,
				rn, cols, &vals[nrows*rs],
				INSERT_VALUES);			CHKERRQ(ierr); 
			ierr = MatAssemblyBegin(A,
				MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
			ierr = MatAssemblyEnd(A,
				MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
			ierr = MatRestoreArray(AT,&vals);	CHKERRQ(ierr);
			PetscFree(rows);
			PetscFree(cols);
		}

		ierr = MatDestroy(AT);
	}
	
	/**** Argument -acre_fb gives the name of the binary file
	 **** that contains the response vector. This vector
	 **** is distributed along each processor column and
	 **** replicated along processor rows.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fb",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fb is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		ierr = MPI_Finalize();
			assert(ierr==MPI_SUCCESS);
		exit(0);
	}
	ierr = acre_read_vector(PETSC_COMM_WORLD,
		infile, format, &b);				CHKERRQ(ierr);

	/**** Argument -acre_fsL gives the name of the binary file
	 **** that contains the sparsity structure of the Cholesky
	 **** factor L of the covariance matrix C=LL'. This vector
	 **** is distributed like the response vector b.
	 **** 
	 **** Note: the structural matrix sL has one row for every
	 **** row of A and L and one column for every (co)variance
	 **** component that is sought, see the above reference 
	 **** paper for details. The following example might help.
	 **** Suppose the exogenous matrix A is 5x2, making the
	 **** matrices C and L to be 5x5, with say 3 components t_1,
	 **** t_2 and t_3 to be estimated. The matrix L and its
	 **** structural matrix sL might look like this with 0
	 **** represented by ".":
	 ****
	 ****        /                     \         /           \              
	 ****        | t_1  .   .   .   .  |         |  1  .  .  |              
	 ****	L =  |                     |   sL =  |           |              
	 ****        | t_3 t_2  .   .   .  |         |  .  2  1  |              
	 ****        |                     |         |           |              
	 ****        | t_3  .  t_1  .   .  |         |  3  .  1  |              
	 ****        |                     |         |           |              
	 ****        | t_3  .   .  t_2  .  |         |  .  4  1  |              
	 ****        |                     |         |           |              
	 ****        | t_3  .   .   .  t_1 |         |  5  .  1  |              
	 ****        \                     /         \           /              
	 ****
	 **** This implementation assumes that no t_i appears more
	 **** than once on any one row of L. That limitation  should
	 **** not be hard to overcome.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fsL",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fsL is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		ierr = MPI_Finalize();
			assert(ierr==MPI_SUCCESS);
		exit(0);
	}

	ierr = acre_read_aij_matrix(PETSC_COMM_WORLD, infile,
			format, &sL);				CHKERRQ(ierr);

	ierr = MatGetSize(sL,&mm, &nq);				CHKERRQ(ierr);
	if ( mm!=m )
	{
		if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
			"Matrix sL has wrong number of rows\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		ierr = MPI_Finalize();
			assert(ierr==MPI_SUCCESS);
		exit(0);
	}

	/**** Argument -acre_fx gives the name of the binary file
	 **** that contains an initial guess on the unknown
	 **** components. If this option is omitted, the algorithm
	 **** starts with all ones, an all zero vector would
	 **** lead to a singular matrix L.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fx",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		ierr = VecCreateSeq ( PETSC_COMM_SELF, nq,
			&cVec);					CHKERRQ(ierr);
		for ( i=0; i<nq; i++ )
		{
			PetscScalar one = 1.0;
			VecSetValue(cVec,i, one, INSERT_VALUES);
		}
		ierr = VecAssemblyBegin( cVec);			CHKERRQ(ierr);
		ierr = VecAssemblyEnd( cVec);			CHKERRQ(ierr);
	}
	else
	{
		/* VecLoad may load the vector as distributed even though
		 * we ask for VECSEQ. It would be nice if the sequential 
		 * vector was replicated across all processes without
		 * a call to acre_bcast_vector. 
		 */
		ierr = acre_read_vector_serial(PETSC_COMM_SELF,
			infile, format, &cVec);			CHKERRQ(ierr);
	}

	/**** Argument -acre_flb gives the name of the binary file
	 **** that contains the lower bounds on the unknown
	 **** components. If this option is omitted, the lower
	 **** bound on all components is zero.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_flb",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		ierr = VecDuplicate(cVec, &lVec );		CHKERRQ(ierr);
		for ( i=0; i<nq; i++ )
		{
			PetscScalar zero = 0.0;
			VecSetValue(lVec,i, zero, INSERT_VALUES);
		}
		ierr = VecAssemblyBegin( lVec);			CHKERRQ(ierr);
		ierr = VecAssemblyEnd( lVec);			CHKERRQ(ierr);
	}
	else
	{
		ierr = acre_read_vector_serial(PETSC_COMM_SELF,
			infile, format, &lVec);			CHKERRQ(ierr);
	}

	/**** Argument -acre_fub gives the name of the binary file
	 **** that contains the lower bounds on the unknown
	 **** components. If this option is omitted, the lower
	 **** bound on all components is zero.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fub",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		ierr = VecDuplicate(cVec, &uVec );		CHKERRQ(ierr);
		for ( i=0; i<nq; i++ )
		{
			PetscScalar huge = 1.0e+32;
			VecSetValue(uVec,i, huge, INSERT_VALUES);
		}
		ierr = VecAssemblyBegin( uVec);			CHKERRQ(ierr);
		ierr = VecAssemblyEnd( uVec);			CHKERRQ(ierr);
	}
	else
	{
		ierr = acre_read_vector_serial(PETSC_COMM_SELF,
			infile, format, &uVec);			CHKERRQ(ierr);
	}

	/**** Argument -acre_ml is set to specify that
	 **** the REML term is not needed.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL,"-acre_ml",
		ftype,8,&flg);					CHKERRQ(ierr);
	if (flg==PETSC_FALSE || !strcasecmp(ftype,"reml"))
	{
		cntxt.ofcn = ACRE_REML;
	} else
	{
		cntxt.ofcn = ACRE_ML;
	}

	/**** Argument -acre_i gives the maximum number of iterations
	 ****/
	ierr = PetscOptionsGetInt(TAO_NULL,"-acre_i",
		&maxits,&flg);					CHKERRQ(ierr);

	/**** Check for -acre_xc although it has already been checked
	 **** for; otherwise PETSc complains about an unused argument.
	 ****/
	{
		int oxc;
	 	ierr = PetscOptionsGetInt( PETSC_NULL,
			"-acre_xc", &oxc, &flg);		CHKERRQ(ierr);
			assert(flg!=PETSC_TRUE || oxc==xc);
	}

	/***
	 *** Bundle all the above into the ACRE context. 
	 ***/
	acre_start( A, b, sL, cVec, xc, &cntxt);

	ierr = acre_set_cholesky(&cntxt, cVec);			CHKERRQ(ierr);

	/***
	 *** Initialize the TAO solver
	 ***/
	ierr = TaoCreate(PETSC_COMM_SELF,"tao_blmvm",
		&reml);						CHKERRQ(ierr);
	ierr = TaoApplicationCreate(PETSC_COMM_SELF,
		&reml_app);					CHKERRQ(ierr);
	ierr = TaoAppSetVariableBounds( reml_app ,
		lVec, uVec);					CHKERRQ(ierr);
	ierr = TaoAppSetObjectiveRoutine( reml_app,
		acre_likelihood, &cntxt);			CHKERRQ(ierr);

	/**** Argument -acre_exact controls the stopping criteria
	 **** of the TAO solver.
	 ****/
	cntxt.tol = 1.0e-15;
	ierr = PetscOptionsGetScalar(TAO_NULL,"-acre_exact",
		&cntxt.tol,&flg);				CHKERRQ(ierr);
	if ( cntxt.tol>0.0 )
	{
		ierr = TaoSetTolerances( reml, cntxt.tol,
			cntxt.tol, TAO_DEFAULT, TAO_DEFAULT);	CHKERRQ(ierr);
		ierr = TaoSetGradientTolerances( reml,
			cntxt.tol, TAO_DEFAULT, TAO_DEFAULT);	CHKERRQ(ierr);
	}
	if(w_rank==0) PetscPrintf(PETSC_COMM_SELF,
		"Tolerance set to %lg\n",cntxt.tol); 

	/**** Argument -acre_g determines which gradient
	 **** formulation will be used. Note that only
	 **** the default (cramer) is available when
	 **** -acre_xc is set larger than 1.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL,
		"-acre_g", gtype, 8, &flg);			CHKERRQ(ierr);
	if (xc>1 || flg!=PETSC_TRUE || !strcasecmp(gtype,
			"cramer"))
	{
		igrad = ACRE_CRAMER_GRADIENT;
		ierr = TaoAppSetGradientRoutine(
			reml_app, acre_cramer_gradient,
			&cntxt);				CHKERRQ(ierr);
		ierr = TaoAppSetObjectiveAndGradientRoutine(
			reml_app, acre_cramer_function_gradient,
			(void *) &cntxt);			CHKERRQ(ierr);
	}
	else
	{
		igrad = ACRE_ADJOINT_GRADIENT;
		ierr = TaoAppSetGradientRoutine( reml_app, 
			acre_adjoint_gradient, &cntxt);	CHKERRQ(ierr);
		ierr = TaoAppSetObjectiveAndGradientRoutine(
			reml_app, acre_adjoint_function_gradient,
			(void *) &cntxt);		CHKERRQ(ierr);
	}
	if(w_rank==0)
		PetscPrintf(PETSC_COMM_SELF,
			"Gradient uses %s formula\n",
			(igrad==ACRE_CRAMER_GRADIENT) ?
			"Cramer's" : "adjoint");

	/**** Argument -acre_vec controls the number of vectors
	 **** used in quasi-Newton methods.
	 ****/
	ierr = PetscOptionsGetInt( PETSC_NULL,
		"-acre_vec", &nv, &flg);		CHKERRQ(ierr);
	if(flg==PETSC_TRUE && nv>=1 )
	{
		ierr = TaoLMVMSetSize( reml, nv);
	}

	ierr = TaoSetMaximumIterates(reml,maxits);	CHKERRQ(ierr);

	/***
	 *** Compute the REML variance components
	 ***/

	ierr = VecSet(&one, cVec);			CHKERRQ(ierr);
	ierr = TaoAppSetInitialSolutionVec(reml_app,
		cVec);					CHKERRQ(ierr);

	ierr = TaoSolveApplication(reml_app, reml);	CHKERRQ(ierr);

	ierr = acre_write_tao_solver_state(reml);	CHKERRQ(ierr);

	PetscViewerFlush(PETSC_VIEWER_STDOUT_SELF);
	PetscViewerFlush(PETSC_VIEWER_STDOUT_WORLD);
	if(w_rank==0)
	{
		PetscPrintf(PETSC_COMM_SELF,
			"Theta=\n");
		ierr = VecView( cVec,
			PETSC_VIEWER_STDOUT_SELF);	CHKERRQ(ierr);
	}
	PetscViewerFlush(PETSC_VIEWER_STDOUT_SELF);

	/* Compute the effects too */
	ierr = VecCreateSeq ( PETSC_COMM_SELF, n, &x);	CHKERRQ(ierr);
	ierr = acre_effects(&cntxt,x,&res);		CHKERRQ(ierr);
	if(w_rank==0)
	{
		PetscPrintf(PETSC_COMM_SELF,
			"x=\n");
		ierr = VecView( x,
			PETSC_VIEWER_STDOUT_SELF);	CHKERRQ(ierr);
		PetscPrintf(PETSC_COMM_SELF,
			"Residual Norm=%lg\n",res);
	}
	PetscViewerFlush(PETSC_VIEWER_STDOUT_SELF);
	ierr = MPI_Barrier(MPI_COMM_WORLD);
		assert(ierr==MPI_SUCCESS);

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	Free any remaining Petsc Object
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	acre_end(&cntxt);
	ierr = TaoDestroy(reml);				CHKERRQ(ierr);
	ierr = TaoAppDestroy(reml_app);				CHKERRQ(ierr);
	ierr = VecDestroy(x);					CHKERRQ(ierr);
	ierr = VecDestroy(b);					CHKERRQ(ierr);
	ierr = MatDestroy(A);					CHKERRQ(ierr);
	
	ierr = MatDestroy(sL);					CHKERRQ(ierr);
	ierr = VecDestroy(cVec);				CHKERRQ(ierr);
	ierr = VecDestroy(lVec);				CHKERRQ(ierr);
	ierr = VecDestroy(uVec);				CHKERRQ(ierr);
	ierr = TaoFinalize();					CHKERRQ(ierr);
	ierr = MPI_Finalize(); 
		assert(ierr==MPI_SUCCESS);
	
	return 0;

}

