/* $Id: acre_mgs.c,v 1.1 2004/04/30 05:18:18 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petsclog.h"
#include "acre.h"
#include "petsclog.h"
#include <assert.h>
#include <mpi.h>
#include <math.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_mgs"
int acre_mgs(acre_cntxt *cntxt, Mat Q, Mat R)
{
	/* Note that both Q and R are dense matrices and R is 
	 * replicated onto all processes.
	 */
	int k, i;
	int nrows, *rows, *cols;
	int start, end;
	int n = cntxt->n;
	int rn = cntxt->rn;
	int rs;
	PetscScalar w, *uals = NULL, *vals=NULL, *wals=NULL;
	PetscScalar *qvals=NULL, *rvals=NULL;
	PetscScalar *qivals, *qjvals;
	int ierr, cank, rank, here;
	int root;
	int icol;
	int ione = 1;
	PetscScalar aone = 1.0;
	PetscScalar amone = -1.0;
	PetscScalar azero = 0.0;
	PetscScalar prs;
	static int ncall;

	HELLO("acre_mgs enter");

	ierr = MPI_Comm_rank(cntxt->rcomm,&rank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&cank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(MPI_COMM_WORLD,&here);
		assert(ierr==MPI_SUCCESS);

	rs = cntxt->rstart[rank];

	ierr = VecGetOwnershipRange( cntxt->b,
		&start, &end );					CHKERRQ(ierr);
	nrows = end-start;
	ierr = PetscMalloc(nrows*sizeof(int),&rows);		CHKERRQ(ierr);
	for ( i=0; i<nrows; i++) rows[i] = i+start;
	ierr = PetscMalloc(rn*sizeof(int),&cols);		CHKERRQ(ierr);
	for ( i=0; i<rn; i++) cols[i] = i+rs;


	if ( cntxt->qc>1 )
	{
		ierr = PetscMalloc(nrows*sizeof(double),&wals);	CHKERRQ(ierr);
	}
	ierr = PetscMalloc(n*sizeof(double),&vals);		CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(double),&uals);		CHKERRQ(ierr);

	/* right looking MGS */
	root = 0;
	icol = 0;
	ierr = MatZeroEntries( R );				CHKERRQ(ierr);
	ierr = MatGetArray(Q,&qvals);
	ierr = MatGetArray(R,&rvals);
	for ( i=0; i<n; i++ )
	{

		/* update root rank if necessary */
		if ( cntxt->rstart[root+1]<=i ) root++;

		/* root process grabs and scales the first active
		 * column of Q.
		 */
		if ( rank==root )
		{
			qivals = qvals+icol*nrows;
			prs = BLdot_(&nrows,qivals,&ione,
				qivals,&ione);
			ierr = MPI_Allreduce(&prs, &w,
				1, MPI_DOUBLE,MPI_SUM,
				PETSC_COMM_WORLD);
				assert(ierr==MPI_SUCCESS);
			w = (w==0) ? 0.0 : sqrt(w);

			/* Complain if R is singular */
			if ( w<1.0 && HUGE*w < 2.0 )
			{
				PetscPrintf(PETSC_COMM_WORLD,
					"MGS failed\n");
				ierr = VecView(cntxt->pVec,
				PETSC_VIEWER_STDOUT_SELF);	CHKERRQ(ierr);
				/* If a flag were set here
				 * the likelihood could be
				 * augmented with a penalty.
				 */
			}

			/* set diagonal term of R */
			ierr = MatSetValues( R, 1, &i, 1,
				&i, &w, INSERT_VALUES);		CHKERRQ(ierr);

			/* scale first active column of Q */
			w = 1.0/w;
			BLscal_(&nrows,&w,qivals,&ione);
			icol++;
		} 
		else
		{
			qivals = wals;
		}

		/* share this column if necessary */
		if ( cntxt->qc>1 )
		{
			ierr = MPI_Bcast(qivals, nrows,
				MPI_DOUBLE, root,
				cntxt->rcomm);
				assert(ierr==MPI_SUCCESS);
		}

		/* update the local Q and R */
		qjvals = qvals+icol*nrows;
		k = rn-icol;
		LAgemv_("T", &nrows, &k, &aone,
			qjvals, &nrows, qivals, &ione,
			&azero, vals, &ione);
		ierr = MPI_Allreduce(vals, uals,
			k, MPI_DOUBLE,MPI_SUM,
			PETSC_COMM_WORLD);
			assert(ierr==MPI_SUCCESS);
		BLger_(&nrows,&k,&amone,qivals,&ione,
			uals,&ione,qjvals,&nrows);

		BLcopy_(&k,uals,&ione,
			&rvals[i+(rs+icol)*n],&n);
	}

	if ( wals!=NULL ) PetscFree( wals );
	PetscFree( vals );
	PetscFree( uals );
	if ( rows!=NULL ) PetscFree( rows );
	if ( cols!=NULL ) PetscFree( cols );

	ierr = MatAssemblyBegin(Q,MAT_FINAL_ASSEMBLY );		CHKERRQ(ierr);
	ierr = MatAssemblyEnd(Q, MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);

	if ( cntxt->qc>1 )
	{
		int n2 = n*n;
		ierr = PetscMalloc(n2*sizeof(double),&vals);	CHKERRQ(ierr);
		BLcopy_(&n2, rvals, &ione, vals, &ione);
		ierr = MPI_Allreduce(vals,rvals,n2,
			MPI_DOUBLE, MPI_SUM, cntxt->rcomm);
			assert(ierr==MPI_SUCCESS);
		PetscFree(vals);
	}

	ierr = MatRestoreArray(Q,&qvals);
	ierr = MatRestoreArray(R,&rvals);

	ierr = MatAssemblyBegin(R,MAT_FINAL_ASSEMBLY );		CHKERRQ(ierr);
	ierr = MatAssemblyEnd(R, MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
	ncall++;

	HELLO("acre_mgs exit");

	return 0;

}
